# Tai Sakuma <tai.sakuma@gmail.com>
import time, random
import itertools
import threading

import pytest

from atpbar import atpbar
from atpbar import flush

##__________________________________________________________________||
@pytest.mark.parametrize('time_starting_task', [0, 0.01, 0.2])
@pytest.mark.parametrize('niterations', [[5, 4, 3], [5, 0, 1], [0], [1]])
@pytest.mark.parametrize('nthreads', [3, 1, 0])
def test_threading_from_loop(mock_create_presentation, wrap_end_pickup, nthreads, niterations, time_starting_task):

    # make niterations as long as nthreads. repeat if necessary
    niterations = list(itertools.chain(*itertools.repeat(niterations, nthreads//len(niterations)+1)))[:nthreads]

    def run_with_threading(nthreads=3, niterations=[5, 5, 5], time_starting_task=0):

        def task(n, name, time_starting):
            time.sleep(time_starting) # When starting time is long, the loop in
                                      # the main thread might already end by the
                                      # time the loop in this task starts.
            for i in atpbar(range(n), name=name):
                time.sleep(0.0001)

        threads = [ ]
        for i in atpbar(range(nthreads)): # `atpbar` is used here while `atpbar`
                                          # is also used in threads being
                                          # launched in this loop. If none of
                                          # the `atpbar`s in threads has started
                                          # by the end of this loop, the
                                          # `atpbar` for this loop waits until
                                          # the progress bar for this loop
                                          # finish updating. Otherwise, progress
                                          # bars from threads are being updated
                                          # together with the progress bar for
                                          # this loop and the `atpbar` will not
                                          # wait.

            name = 'thread {}'.format(i)
            n = niterations[i]
            t = threading.Thread(target=task, args=(n, name, time_starting_task))
            t.start()
            threads.append(t)

            time.sleep(0.01)
            # sleep sometime so this loop doesn't end too quickly. Without this
            # sleep, this loop could end before an `atpbar` in any of the
            # threads start even if `time_starting_task` is zero.

        for t in threads:
            t.join()

        flush()

    run_with_threading(nthreads, niterations, time_starting_task)

    ## print()
    ## print(mock_create_presentation)

    ## assert 1 == wrap_end_pickup.call_count # 1 or 2 depending on how quickly the loop ends

    nreports_expected_from_main = nthreads + 1
    nreports_expected_from_threads = sum(niterations) + nthreads
    nreports_expected = nreports_expected_from_main + nreports_expected_from_threads

    presentations = mock_create_presentation.presentations

    if nreports_expected_from_threads == 0:
        assert 3 == len(presentations) # when `atpbar` in the main thread
                                       # starts and end and when flush() is
                                       # called

        progressbar0 = presentations[0]
        assert nreports_expected == len(progressbar0.reports)
        # one report from `atpbar` in the main thread

        assert 1 == len(progressbar0.taskids)
        assert 1 == progressbar0.nfirsts
        assert 1 == progressbar0.nlasts

    else:

        if 2 == len(presentations):

            progressbar0 = presentations[0]
            assert nreports_expected == len(progressbar0.reports)
            assert nthreads + 1 == len(progressbar0.taskids)
            assert nthreads + 1 == progressbar0.nfirsts
            assert nthreads + 1 == progressbar0.nlasts

            progressbar1 = presentations[1]
            assert 0 == len(progressbar1.reports)

        else:
            assert 3 == len(presentations)

            progressbar0 = presentations[0]
            progressbar1 = presentations[1]

            assert nreports_expected == len(progressbar0.reports) + len(progressbar1.reports)
            assert nthreads + 1 == len(progressbar0.taskids) + len(progressbar1.taskids)
            assert nthreads + 1 == progressbar0.nfirsts + progressbar1.nfirsts
            assert nthreads + 1 == progressbar0.nlasts + progressbar1.nlasts

            progressbar2 = presentations[2]
            assert 0 == len(progressbar2.reports)

    # At this point the pickup shouldn't be owned. Therefore, a new
    # `atpbar` in the main thread should own it.
    npresentations = len(presentations)
    for i in atpbar(range(4)):
        pass
    assert npresentations + 1 == len(presentations)
    progressbar = presentations[-2]
    assert 4 + 1 == len(progressbar.reports)
    assert 1 == len(progressbar.taskids)
    assert 1 == progressbar.nfirsts
    assert 1 == progressbar.nlasts

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os, uuid
import logging

import contextlib

from .funcs import fetch_reporter, in_main_thread

##__________________________________________________________________||
def atpbar(iterable, name=None):
    """returns an instance of `Atpbar`

    Parameters
    ----------
    iterable : iterable
        An iterable whose progress of the iterations is to be shown
    name : str
        A label to be shown on the progress bar

    Returns
    -------
    iterable
        An instance of `Atpbar` if successfully instantiated.
        Otherwise, the object received as the parameter `iterable`.

    """
    try:
        len_ = len(iterable)
    except TypeError:
        logger = logging.getLogger(__name__)
        logging.warning('length is unknown: {!r}'.format(iterable))
        logging.warning('atpbar is turned off')
        return iterable

    if name is None:
        name = repr(iterable)

    return Atpbar(iterable, name=name, len_=len_)

##__________________________________________________________________||
class Atpbar(object):
    """Progress bar

    An iterable that wraps another iterable and shows the progress
    bars for the iterations. The class is usually instantiated by the
    function `atpbar`.

    Parameters
    ----------
    iterable : iterable
        An iterable whose progress of the iterations is to be shown
    name : str
        A label to be shown on the progress bar
    len_ : int
        The length of the iterable

    """

    def __init__(self, iterable, name, len_):
        self.iterable = iterable
        self.name = name
        self.len_ = len_
        self.id_ = uuid.uuid4()

    def __iter__(self):
        with fetch_reporter() as reporter:
            self.reporter = reporter
            self.loop_complete = False
            self._report_start()
            with report_last(pbar=self):
                for i, e in enumerate(self. iterable):
                    yield e
                    self._report_progress(i)
                else:
                    self.loop_complete = True

    def _report_start(self):
        if self.reporter is None:
            return
        try:
            report = dict(
                taskid=self.id_, name=self.name,
                done=0, total=self.len_,
                pid=os.getpid(), in_main_thread=in_main_thread())
            self.reporter.report(report)
        except:
            pass

    def _report_progress(self, i):
        if self.reporter is None:
            return
        try:
            report = dict(taskid=self.id_, done=(i+1))
            self.reporter.report(report)
        except:
            pass

@contextlib.contextmanager
def report_last(pbar):
    """send a last report

    This function sends the last report of the task when the loop ends
    with `break` or an exception so that the progress bar will be
    updated with the last complete iteration.

    """
    try:
        yield
    finally:
        if not pbar.loop_complete:
            if pbar.reporter is not None:
                try:
                    report = dict(taskid=pbar.id_, first=False, last=True)
                    pbar.reporter.report(report)
                except:
                    pass

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
from .report import ProgressReportComplementer

##__________________________________________________________________||
DEFAULT_INTERVAL = 0.1 # [second]

##__________________________________________________________________||
class ProgressReporter(object):
    """A progress reporter

    This class sends progress reports. The reports will be picked up
    by the pickup (`ProgressReportPickup`), which uses the reports,
    for example, to update `ProgressBar` on the screen.

    An instance of this class is initialized with a message queue::

        reporter = ProgressReporter(queue)

    The pickup, which is running in a sub-thread of the main process,
    needs to have the same queue.

    A report can be sent as::

        reporter.report(report)

    This method can be frequently called multiple times. However,
    after sending one report, the reporter wait for a certain
    ``interval`` (0.1 seconds by default) before sending another
    report of the same task. Reports from the same task received
    within this interval will be discarded. The exception for this is
    the last report. The last report, which indicates the completion
    of the task, will be always sent to the progress monitor
    regardless of whether it is given within the interval.

    """
    def __init__(self, queue):
        self.queue = queue
        self.interval = DEFAULT_INTERVAL # [second]
        self.last_time = { } # key: taskid
        self.complete_report = ProgressReportComplementer()

    def __repr__(self):
        return '{}(queue={!r}, interval={!r})'.format(
            self.__class__.__name__,
            self.queue,
            self.interval
        )

    def report(self, report):
        """send ``report`` to a progress monitor

        Args:
            report (ProgressReport): a progress report

        """

        self.complete_report(report)

        if not self._need_to_report(report):
            return

        self.queue.put(report)

        self.last_time[report['taskid']] = self._time()

    def _need_to_report(self, report):

        if report['first']:
            return True

        if report['last']:
            return True

        if report['taskid'] not in self.last_time:
            return True

        if self._time() - self.last_time[report['taskid']] > self.interval:
            return True

        return False

    def _time(self):
        return time.time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from atpbar.presentation.txtprint import ProgressPrint

##__________________________________________________________________||
@pytest.fixture()
def mock_time(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(time, 'time', ret)
    ret.return_value = 1533374055.904203
    return ret

@pytest.fixture()
def obj(mock_time):
    return ProgressPrint()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

def test_report(obj, capsys):
    report = dict(taskid=1, name='task1', done=0, total=10, first=True, last=False)
    obj.present(report)
    captured = capsys.readouterr()
    stdout_lines = captured.out.strip().split('\n')
    assert 1 == len(stdout_lines)
    assert (' :        0 /       10 (  0.00%): task1' in stdout_lines[0])

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys

from .bartty import ProgressBar
from .txtprint import ProgressPrint

try:
    from .barjupyter import ProgressBarJupyter
except ImportError:
    pass

##__________________________________________________________________||
def create_presentation():
    if sys.stdout.isatty():
        return ProgressBar()
    if is_jupyter_notebook():
        return ProgressBarJupyter()
    return ProgressPrint()

def is_jupyter_notebook():
    try:
        from IPython import get_ipython
        if 'IPKernelApp' in get_ipython().config:
            return True
    except:
        pass
    return False

##__________________________________________________________________||
